import cv2
import numbers
import random
from .cvEnhance import Color, Brightness, Contrast
import numpy as np


class Lambda(object):
    """Apply a user-defined lambda as a transform.

    Args:
        lambd (function): Lambda/function to be used for transform.
    """

    def __init__(self, lambd):
        assert callable(lambd), repr(type(lambd).__name__) + " object is not callable"
        self.lambd = lambd

    def __call__(self, img):
        return self.lambd(img)

    def __repr__(self):
        return self.__class__.__name__ + '()'

def _is_np_image(img):
    return type(img)==np.ndarray

def adjust_brightness(img, brightness_factor):
    """Adjust brightness of an Image.

    Args:
        img (np.unit8): numpy ndarray with BGR channel order.
        brightness_factor (float):  How much to adjust the brightness. Can be
            any non negative number. 0 gives a black image, 1 gives the
            original image while 2 increases the brightness by a factor of 2.

    Returns:
        numpy ndarray: Brightness adjusted image.
    """
    if not _is_np_image(img):
        raise TypeError('img should be numpy ndarray. Got {}'.format(type(img)))
    enhancer = Brightness(img)
    img = enhancer.enhance(brightness_factor)
    return img

def adjust_contrast(img, contrast_factor):
    """Adjust contrast of an Image.

    Args:
        img (np.unit8): numpy ndarray with BGR channel order.
        contrast_factor (float): How much to adjust the contrast. Can be any
            non negative number. 0 gives a solid gray image, 1 gives the
            original image while 2 increases the contrast by a factor of 2.

    Returns:
        numpy ndarray: Contrast adjusted image.
    """
    if not _is_np_image(img):
        raise TypeError('img should be numpy ndarray. Got {}'.format(type(img)))
    enhancer = Contrast(img)
    img = enhancer.enhance(contrast_factor)
    return img

def adjust_saturation(img, saturation_factor):
    """Adjust color saturation of an image.
    Args:
        img (np.unit8): numpy ndarray with BGR channel order.
        saturation_factor (float):  How much to adjust the saturation. 0 will
            give a black and white image, 1 will give the original image while
            2 will enhance the saturation by a factor of 2.

    Returns:
        numpy ndarray: Saturation adjusted image.
    """
    if not _is_np_image(img):
        raise TypeError('img should be numpy ndarray. Got {}'.format(type(img)))
    enhancer = Color(img)
    img = enhancer.enhance(saturation_factor)
    return img

def adjust_hue(img, hue_factor):
    """Adjust hue of an image.

    The image hue is adjusted by converting the image to HSV and
    cyclically shifting the intensities in the hue channel (H).
    The image is then converted back to original image mode.

    `hue_factor` is the amount of shift in H channel and must be in the
    interval `[-0.5, 0.5]`.

    See `Hue`_ for more details.

    .. _Hue: https://en.wikipedia.org/wiki/Hue

    Args:
        img (np.unit8): numpy ndarray with BGR channel order.
        hue_factor (float):  How much to shift the hue channel. Should be in
            [-0.5, 0.5]. 0.5 and -0.5 give complete reversal of hue channel in
            HSV space in positive and negative direction respectively.
            0 means no shift. Therefore, both -0.5 and 0.5 will give an image
            with complementary colors while 0 gives the original image.

    Returns:
        numpy ndarray: Hue adjusted image.
    """
    if not _is_np_image(img):
        raise TypeError('img should be numpy ndarray. Got {}'.format(type(img)))
    if not(-0.5 <= hue_factor <= 0.5):
        raise ValueError('hue_factor is not in [-0.5, 0.5].'.format(hue_factor))
    imghsv = img.copy().astype(np.uint8)
    imghsv = cv2.cvtColor(imghsv, cv2.COLOR_BGR2HSV)
    # note that the hue value range from 0 to 180, (hue in [0,180) )
    add_on = np.uint8(hue_factor * 180)
    with np.errstate(over='ignore'):
        imghsv[:,:,0] += add_on
    return cv2.cvtColor(imghsv, cv2.COLOR_HSV2BGR)

class ColorJitter(object):
    def __init__(self, brightness=0, contrast=0, saturation=0, hue=0):
        def _check_input(value, name, center=1, bound=(0, float('inf')), clip_first_on_zero=True):
            if isinstance(value, numbers.Number):
                if value < 0:
                    raise ValueError("If {} is a single number, it must be non negative.".format(name))
                value = [center - value, center + value]
                if clip_first_on_zero:
                    value[0] = max(value[0], 0)
            elif isinstance(value, (tuple, list)) and len(value) == 2:
                if not bound[0] <= value[0] <= value[1] <= bound[1]:
                    raise ValueError("{} values should be between {}".format(name, bound))
            else:
                raise TypeError("{} should be a single number or a list/tuple with lenght 2.".format(name))

            # if value is 0 or (1., 1.) for brightness/contrast/saturation
            # or (0., 0.) for hue, do nothing
            if value[0] == value[1] == center:
                value = None
            return value
        self.brightness = _check_input(brightness, 'brightness')
        self.contrast = _check_input(contrast, 'contrast')
        self.saturation = _check_input(saturation, 'saturation')
        self.hue = _check_input(hue, 'hue', center=0, bound=(-0.5, 0.5),
                                     clip_first_on_zero=False)
        print(self.brightness)
    def __call__(self, img):
        img = img.astype(np.uint8)
        transforms = []
        brightness, contrast, saturation, hue = \
                self.brightness, self.contrast, self.saturation, self.hue
        
        if brightness is not None:
            brightness_factor = random.uniform(brightness[0], brightness[1])
            transforms.append(Lambda(lambda img: adjust_brightness(img, brightness_factor)))

        if contrast is not None:
            contrast_factor = random.uniform(contrast[0], contrast[1])
            transforms.append(Lambda(lambda img: adjust_contrast(img, contrast_factor)))

        if saturation is not None:
            saturation_factor = random.uniform(saturation[0], saturation[1])
            transforms.append(Lambda(lambda img: adjust_saturation(img, saturation_factor)))

        if hue is not None:
            hue_factor = random.uniform(hue[0], hue[1])
            transforms.append(Lambda(lambda img: adjust_hue(img, hue_factor)))

        random.shuffle(transforms)
        for t in transforms:
            img = t(img)
        return img

    def __repr__(self):
        format_string = self.__class__.__name__ + '('
        format_string += 'brightness={0}'.format(self.brightness)
        format_string += ', contrast={0}'.format(self.contrast)
        format_string += ', saturation={0}'.format(self.saturation)
        format_string += ', hue={0})'.format(self.hue)
        return format_string



import cv2
import numpy as np

def blend(im1, im2, alpha):
    return (im1 * (1 - alpha) + im2 * alpha).astype(np.uint8)

class _Enhance(object):
    def enhance(self, factor):
        """
        Returns an enhanced image.
        :param factor: A floating point value controlling the enhancement.
                       Factor 1.0 always returns a copy of the original image,
                       lower factors mean less color (brightness, contrast,
                       etc), and higher values more. There are no restrictions
                       on this value.
        :rtype: numpy array(image)
        """
        return blend(self.degenerate, self.image, factor)

class Contrast(_Enhance):
    def __init__(self, image):
        self.image = image.copy()
        mean = int(cv2.cvtColor(image,cv2.COLOR_BGR2GRAY).mean() + 0.5)
        imgd = (np.ones(image.shape[:2]) * mean).astype(np.uint8)
        self.degenerate = cv2.cvtColor(imgd, cv2.COLOR_GRAY2BGR)

class Brightness(_Enhance):
    def __init__(self, image):
        self.image = image.copy()
        self.degenerate = np.zeros(image.shape)

class Color(_Enhance):
    def __init__(self, image):
        self.image = image.copy()
        gray = cv2.cvtColor(image,cv2.COLOR_BGR2GRAY)
        self.degenerate = np.stack([gray]*3, axis=2)



import torch

FLIP_LEFT_RIGHT = 0
FLIP_TOP_BOTTOM = 1

class BoxList(object):
    def __init__(self, bbox, image_size, mode='xyxy', device=torch.device('cpu')):
        x = torch.as_tensor(bbox, dtype=torch.float32, device=device)
        if bbox.ndimension() != 2:
            raise ValueError(
                "bbox should have 2 dimensions, got {}".format(bbox.ndimension())
            )
        if bbox.size(-1) != 4:
            raise ValueError(
                "last dimension of bbox should have a "
                "size of 4, got {}".format(bbox.size(-1))
            )
        if mode not in ("xyxy", "xywh"):
            raise ValueError("mode should be 'xyxy' or 'xywh'")

        self.bbox = bbox
        self.size = image_size
        self.mode = mode 
        self.extra_fields = {}

    def add_field(self, field, field_data):
        self.extra_fields[field] = field_data

    def get_field(self, field):
        return self.extra_fields[field]

    def has_field(self, field):
        return field in self.extra_fields

    def fields(self):
        return list(self.extra_fields.keys())

    def _copy_extra_fields(self, bbox):
        for k, v in bbox.extra_fields.items():
            self.extra_fields[k] = v

    def _split_into_xyxy(self):
        if self.mode == "xyxy":
            xmin, ymin, xmax, ymax = self.bbox.split(1, dim=-1)
            return xmin, ymin, xmax, ymax
        elif self.mode == "xywh":
            xmin, ymin, w, h = self.bbox.split(1, dim=-1)
            return (
                xmin,
                ymin,
                xmin + w.clamp(min=0),
                ymin + h.clamp(min=0),
            )
        else:
            raise RuntimeError("Should not be here")


    def convert(self, mode):
        if mode not in ("xyxy", "xywh"):
            raise ValueError("mode should be 'xyxy' or 'xywh'")
        if mode == self.mode:
            return self
        # we only have two modes, so don't need to check
        # self.mode
        xmin, ymin, xmax, ymax = self._split_into_xyxy()
        if mode == "xyxy":
            bbox = torch.cat((xmin, ymin, xmax, ymax), dim=-1)
            bbox = BoxList(bbox, self.size, mode=mode)
        else:
            bbox = torch.cat(
                (xmin, ymin, xmax - xmin, ymax - ymin), dim=-1
            )
            bbox = BoxList(bbox, self.size, mode=mode)
        bbox._copy_extra_fields(self)
        return bbox

    def resize(self, size, *args, **kwargs):
        """
        Returns a resized copy of this bounding box

        :param size: The requested size in pixels, as a 2-tuple:
            (width, height).
        """

        ratios = tuple(float(s) / float(s_orig) for s, s_orig in zip(size, self.size))
        if ratios[0] == ratios[1]:
            ratio = ratios[0]
            scaled_box = self.bbox * ratio
            bbox = BoxList(scaled_box, size, mode=self.mode)
            # bbox._copy_extra_fields(self)
            for k, v in self.extra_fields.items():
                if not isinstance(v, torch.Tensor):
                    v = v.resize(size, *args, **kwargs)
                bbox.add_field(k, v)
            return bbox

        ratio_width, ratio_height = ratios
        xmin, ymin, xmax, ymax = self._split_into_xyxy()
        scaled_xmin = xmin * ratio_width
        scaled_xmax = xmax * ratio_width
        scaled_ymin = ymin * ratio_height
        scaled_ymax = ymax * ratio_height
        scaled_box = torch.cat(
            (scaled_xmin, scaled_ymin, scaled_xmax, scaled_ymax), dim=-1
        )
        bbox = BoxList(scaled_box, size, mode="xyxy")
        # bbox._copy_extra_fields(self)
        for k, v in self.extra_fields.items():
            if not isinstance(v, torch.Tensor):
                v = v.resize(size, *args, **kwargs)
            bbox.add_field(k, v)

        return bbox.convert(self.mode)

    def transpose(self, method):
        if method not in (FLIP_LEFT_RIGHT, FLIP_TOP_BOTTOM):
            raise NotImplementedError(
                "Only FLIP_LEFT_RIGHT and FLIP_TOP_BOTTOM implemented"
            )

        image_width, image_height = self.size
        xmin, ymin, xmax, ymax = self._split_into_xyxy()
        if method == FLIP_LEFT_RIGHT:
            TO_REMOVE = 1
            transposed_xmin = image_width - xmax - TO_REMOVE
            transposed_xmax = image_width - xmin - TO_REMOVE
            transposed_ymin = ymin
            transposed_ymax = ymax
            transposed_size = self.size
        elif method == FLIP_TOP_BOTTOM:
            transposed_xmin = xmin
            transposed_xmax = xmax
            transposed_ymin = image_height - ymax
            transposed_ymax = image_height - ymin
            transposed_size = self.size
        else:
            RuntimeError("Should not be here")

        bbox = BoxList(transposed_boxes, transposed_size, mode="xyxy")
        # bbox._copy_extra_fields(self)
        for k, v in self.extra_fields.items():
            if not isinstance(v, torch.Tensor):
                v = v.transpose(method)
            bbox.add_field(k, v)
        return bbox.convert(self.mode)

    def to(self, device):
        bbox = BoxList(self.bbox.to(device), self.size, self.mode)
        for k, v in self.extra_fields.items():
            if hasattr(v, "to"):
                v = v.to(device)
            bbox.add_field(k, v)
        return bbox

    def __getitem__(self, item):
        bbox = BoxList(self.bbox[item], self.size, self.mode)
        for k, v in self.extra_fields.items():
            bbox.add_field(k, v[item])
        return bbox

    def __len__(self):
        return self.bbox.shape[0]

    def clip_to_image(self, remove_empty=True):
        TO_REMOVE = 1
        self.bbox[:, 0].clamp_(min=0, max=self.size[0] - TO_REMOVE)
        self.bbox[:, 1].clamp_(min=0, max=self.size[1] - TO_REMOVE)
        self.bbox[:, 2].clamp_(min=0, max=self.size[0] - TO_REMOVE)
        self.bbox[:, 3].clamp_(min=0, max=self.size[1] - TO_REMOVE)
        if remove_empty:
            box = self.bbox
            keep = (box[:, 3] > box[:, 1]) & (box[:, 2] > box[:, 0])
            return self[keep]
        return self

    def area(self):
        box = self.bbox
        if self.mode == "xyxy":
            TO_REMOVE = 1
            area = (box[:, 2] - box[:, 0] + TO_REMOVE) * (box[:, 3] - box[:, 1] + TO_REMOVE)
        elif self.mode == "xywh":
            area = box[:, 2] * box[:, 3]
        else:
            raise RuntimeError("Should not be here")

        return area

    def copy_with_fields(self, fields, skip_missing=False):
        bbox = BoxList(self.bbox, self.size, self.mode)
        if not isinstance(fields, (list, tuple)):
            fields = [fields]
        for field in fields:
            if self.has_field(field):
                bbox.add_field(field, self.get_field(field))
            elif not skip_missing:
                raise KeyError("Field '{}' not found in {}".format(field, self))
        return bbox

    def __repr__(self):
        s = self.__class__.__name__ + "("
        s += "num_boxes={}, ".format(len(self))
        s += "image_width={}, ".format(self.size[0])
        s += "image_height={}, ".format(self.size[1])
        s += "mode={})".format(self.mode)
        return s




from setuptools import setup, find_packages
import sys, os

with open('README.md', 'rb') as f:
    readme = f.read().decode('utf-8')

install_requires = []

with open('requirements.txt', 'rb') as f:
    for req in f.readlines():
        install_requires.append(req.strip())

setup(
    name='cvtorch',
    version='0.0.1',
    description='vision tools based on opencv',
    long_description=readme,
    author='iHateTa11B0y',
    author_email='1187203155@qq.com',
    install_requires=[], #install_requires,
    packages=["cvtorch"],
)
